
VER_NT_WORKSTATION = 0x0000001
VER_NT_DOMAIN_CONTROLLER = 0x0000002
VER_NT_SERVER = 0x0000003
VER_SERVER_NT = 0x80000000
VER_WORKSTATION_NT = 0x40000000
VER_SUITE_SMALLBUSINESS = 0x00000001
VER_SUITE_ENTERPRISE = 0x00000002
VER_SUITE_BACKOFFICE = 0x00000004
VER_SUITE_COMMUNICATIONS = 0x00000008
VER_SUITE_TERMINAL = 0x00000010
VER_SUITE_SMALLBUSINESS_RESTRICTED = 0x00000020
VER_SUITE_EMBEDDEDNT = 0x00000040
VER_SUITE_DATACENTER = 0x00000080
VER_SUITE_SINGLEUSERTS = 0x00000100
VER_SUITE_PERSONAL = 0x00000200
VER_SUITE_BLADE = 0x00000400
VER_SUITE_EMBEDDED_RESTRICTED = 0x00000800
VER_SUITE_SECURITY_APPLIANCE = 0x00001000
VER_SUITE_STORAGE_SERVER = 0x00002000
VER_SUITE_COMPUTE_SERVER = 0x00004000
VER_SUITE_WH_SERVER = 0x00008000

PROCESSOR_ARCHITECTURE_AMD64 = 9
PROCESSOR_ARCHITECTURE_IA64 = 6
PROCESSOR_ARCHITECTURE_INTEL = 0
PROCESSOR_ARCHITECTURE_UNKNOWN = 0xffff

PRODUCT_BUSINESS = 0x00000006
PRODUCT_BUSINESS_N = 0x00000010
PRODUCT_CLUSTER_SERVER = 0x00000012
PRODUCT_DATACENTER_SERVER = 0x00000008
PRODUCT_DATACENTER_SERVER_CORE = 0x0000000C
PRODUCT_DATACENTER_SERVER_CORE_V = 0x00000027
PRODUCT_DATACENTER_SERVER_V = 0x00000025
PRODUCT_ENTERPRISE = 0x00000004
PRODUCT_ENTERPRISE_E = 0x00000046
PRODUCT_ENTERPRISE_N = 0x0000001B
PRODUCT_ENTERPRISE_SERVER = 0x0000000A
PRODUCT_ENTERPRISE_SERVER_CORE = 0x0000000E
PRODUCT_ENTERPRISE_SERVER_CORE_V = 0x00000029
PRODUCT_ENTERPRISE_SERVER_IA64 = 0x0000000F
PRODUCT_ENTERPRISE_SERVER_V = 0x00000026
PRODUCT_HOME_BASIC = 0x00000002
PRODUCT_HOME_BASIC_E = 0x00000043
PRODUCT_HOME_BASIC_N = 0x00000005
PRODUCT_HOME_PREMIUM = 0x00000003
PRODUCT_HOME_PREMIUM_E = 0x00000044
PRODUCT_HOME_PREMIUM_N = 0x0000001A
PRODUCT_HYPERV = 0x0000002A
PRODUCT_MEDIUMBUSINESS_SERVER_MANAGEMENT = 0x0000001E
PRODUCT_MEDIUMBUSINESS_SERVER_MESSAGING = 0x00000020
PRODUCT_MEDIUMBUSINESS_SERVER_SECURITY = 0x0000001F
PRODUCT_PROFESSIONAL = 0x00000030
PRODUCT_PROFESSIONAL_E = 0x00000045
PRODUCT_PROFESSIONAL_N = 0x00000031
PRODUCT_SERVER_FOR_SMALLBUSINESS = 0x00000018
PRODUCT_SERVER_FOR_SMALLBUSINESS_V = 0x00000023
PRODUCT_SERVER_FOUNDATION = 0x00000021
PRODUCT_SMALLBUSINESS_SERVER = 0x00000009
PRODUCT_STANDARD_SERVER = 0x00000007
PRODUCT_STANDARD_SERVER_CORE = 0x0000000D
PRODUCT_STANDARD_SERVER_CORE_V = 0x00000028
PRODUCT_STANDARD_SERVER_V = 0x00000024
PRODUCT_STARTER = 0x0000000B
PRODUCT_STARTER_E = 0x00000042
PRODUCT_STARTER_N = 0x0000002F
PRODUCT_STORAGE_ENTERPRISE_SERVER = 0x00000017
PRODUCT_STORAGE_EXPRESS_SERVER = 0x00000014
PRODUCT_STORAGE_STANDARD_SERVER = 0x00000015
PRODUCT_STORAGE_WORKGROUP_SERVER = 0x00000016
PRODUCT_UNDEFINED = 0x00000000
PRODUCT_ULTIMATE = 0x00000001
PRODUCT_ULTIMATE_E = 0x00000047
PRODUCT_ULTIMATE_N = 0x0000001C
PRODUCT_WEB_SERVER = 0x00000011
PRODUCT_WEB_SERVER_CORE = 0x0000001D

PRODUCT_SUITE_BUSINESS = [PRODUCT_BUSINESS, PRODUCT_BUSINESS_N]
PRODUCT_SUITE_CLUSTER = [PRODUCT_CLUSTER_SERVER]
PRODUCT_SUITE_DATACENTER = [PRODUCT_DATACENTER_SERVER, PRODUCT_DATACENTER_SERVER_CORE, \
                            PRODUCT_DATACENTER_SERVER_CORE_V, PRODUCT_DATACENTER_SERVER_V]
PRODUCT_SUITE_ENTERPRISE = [PRODUCT_ENTERPRISE, PRODUCT_ENTERPRISE_E, PRODUCT_ENTERPRISE_N, \
                            PRODUCT_ENTERPRISE_SERVER, PRODUCT_ENTERPRISE_SERVER_CORE, \
                            PRODUCT_ENTERPRISE_SERVER_CORE_V, PRODUCT_ENTERPRISE_SERVER_IA64, \
                            PRODUCT_ENTERPRISE_SERVER_V]
PRODUCT_SUITE_HOME_BASIC = [PRODUCT_HOME_BASIC, PRODUCT_HOME_BASIC_E, PRODUCT_HOME_BASIC_N]
PRODUCT_SUITE_HOME_PREMIUM = [PRODUCT_HOME_PREMIUM, PRODUCT_HOME_BASIC_E, PRODUCT_HOME_PREMIUM_N]
PRODUCT_SUITE_SMALL_BUSINESS = [PRODUCT_MEDIUMBUSINESS_SERVER_MANAGEMENT, \
                                PRODUCT_MEDIUMBUSINESS_SERVER_MESSAGING, \
                                PRODUCT_MEDIUMBUSINESS_SERVER_SECURITY, \
                                PRODUCT_SMALLBUSINESS_SERVER, PRODUCT_SERVER_FOR_SMALLBUSINESS, \
                                PRODUCT_SERVER_FOR_SMALLBUSINESS_V]
PRODUCT_SUITE_PROFESSIONAL = [PRODUCT_PROFESSIONAL, PRODUCT_PROFESSIONAL_E, PRODUCT_PROFESSIONAL_N]
PRODUCT_SUITE_STANDARD = [PRODUCT_STANDARD_SERVER, PRODUCT_STANDARD_SERVER_CORE, \
                          PRODUCT_STANDARD_SERVER_CORE_V, PRODUCT_STANDARD_SERVER_V]
PRODUCT_SUITE_STARTER = [PRODUCT_STARTER, PRODUCT_STARTER_E, PRODUCT_STARTER_N]
PRODUCT_SUITE_STORAGE = [PRODUCT_STORAGE_ENTERPRISE_SERVER, PRODUCT_STORAGE_EXPRESS_SERVER, \
                         PRODUCT_STORAGE_STANDARD_SERVER, PRODUCT_STORAGE_WORKGROUP_SERVER]
PRODUCT_SUITE_ULTIMATE = [PRODUCT_ULTIMATE, PRODUCT_ULTIMATE_E, PRODUCT_ULTIMATE_N]
PRODUCT_SUITE_WEB = [PRODUCT_WEB_SERVER, PRODUCT_WEB_SERVER_CORE]

SERVER_CORE = [PRODUCT_DATACENTER_SERVER_CORE, PRODUCT_DATACENTER_SERVER_CORE_V, \
               PRODUCT_ENTERPRISE_SERVER_CORE, PRODUCT_ENTERPRISE_SERVER_CORE_V, \
               PRODUCT_STANDARD_SERVER_CORE, PRODUCT_STANDARD_SERVER_CORE_V, \
               PRODUCT_WEB_SERVER_CORE]

HYPER_V = [PRODUCT_DATACENTER_SERVER_CORE_V, PRODUCT_DATACENTER_SERVER_V, \
           PRODUCT_ENTERPRISE_SERVER_CORE_V, PRODUCT_ENTERPRISE_SERVER_V, \
           PRODUCT_SERVER_FOR_SMALLBUSINESS_V, PRODUCT_STANDARD_SERVER_CORE_V, \
           PRODUCT_STANDARD_SERVER_V, PRODUCT_STANDARD_SERVER_CORE_V, \
           PRODUCT_STANDARD_SERVER_V]




from infi.cwrap import WrappedFunction, errcheck_zero, errcheck_nothing, IN, IN_OUT
from ctypes import c_void_p, c_buffer, c_ulong, byref, create_string_buffer

class LibraryFunction(WrappedFunction):
    @classmethod
    def _get_library(cls):
        try:
            import ctypes
            return ctypes.windll.kernel32
        except:
            raise OSError

class GetVersionExA(LibraryFunction):
    @classmethod
    def get_errcheck(cls):
        return errcheck_zero()

    @classmethod
    def get_parameters(cls):
        return ((c_void_p, IN_OUT, "lpVersionInfo"),)

class GetSystemInfo(LibraryFunction):
    @classmethod
    def get_errcheck(cls):
        return errcheck_nothing()

    @classmethod
    def get_parameters(cls):
        return ((c_void_p, IN_OUT, "lpSystemInfo"),)

class GetProductInfo(LibraryFunction):
    @classmethod
    def get_errcheck(cls):
        return errcheck_zero()

    @classmethod
    def get_parameters(cls):
        return ((c_ulong, IN, "majorVersion"),
                (c_ulong, IN, "minorVersion"),
                (c_ulong, IN, "spMajorVersion"),
                (c_ulong, IN, "spMinorVersion"),
                (c_void_p, IN_OUT, "returnedProductType"))

def get_version_ex():
    from .structures import OSVersionEx
    instance = OSVersionEx()
    instance.version_info_size = OSVersionEx.min_max_sizeof().max #pylint: disable-msg=E1101
    instance.csd_version = '\x00' * 128
    buff = create_string_buffer(OSVersionEx.write_to_string(instance)) #pylint: disable-msg=E1101
    GetVersionExA(buff)
    return OSVersionEx.create_from_string(buff) #pylint: disable-msg=E1101

def get_system_info():
    from .structures import SystemInfo
    buff = c_buffer(SystemInfo.min_max_sizeof().max) #pylint: disable-msg=E1101
    GetSystemInfo(buff)
    return SystemInfo.create_from_string(buff) #pylint: disable-msg=E1101

def get_product_info(major_version, minor_version, service_pack_major, service_pack_minor):
    result = c_ulong()
    GetProductInfo(major_version, minor_version, service_pack_major, service_pack_minor, byref(result))
    return result.value




from infi.instruct import Struct, ULInt32, FixedSizeString, ULInt16, ULInt8, ULInt64

def _is_64bit():
    from sys import maxsize
    return maxsize > 2 ** 32

DWORD = ULInt32
TCHAR = FixedSizeString
WORD = ULInt16
BYTE = ULInt8
LPVOID = ULInt64 if _is_64bit() else ULInt32
BOOL = DWORD

class OSVersionEx(Struct):
    _fields_ = [
        DWORD("version_info_size", 0),
        DWORD("major_version", 0),
        DWORD("minor_version", 0),
        DWORD("build_number", 0),
        DWORD("platform_id", 0),
        TCHAR("csd_version", 128),
        WORD("service_pack_major", 0),
        WORD("service_pack_minor", 0),
        WORD("suite_mask", 0),
        BYTE("product_type", 0),
        BYTE("reserved", 0)
        ]

class SystemInfo(Struct):
    _fields_ = [
        WORD("processor_architecture"),
        WORD("reserved"),
        DWORD("page_Size"),
        LPVOID("minimum_application_address"),
        LPVOID("maximum_application_address"),
        DWORD("active_processor_mask"),
        DWORD("number_of_processors"),
        DWORD("processor_type"),
        DWORD("allocation_granularity"),
        WORD("processor_level"),
        WORD("processor_revision")
        ]



__import__("pkg_resources").declare_namespace(__name__)

version_to_name = {
    (5, 0): 'Windows 2000',
    (5, 1): 'Windows XP',
    (5, 2): 'Windows Server 2003',
    (6, 0, 1): 'Windows Vista',
    (6, 0): 'Windows Server 2008',
    (6, 1, 1): 'Windows 7',
    (6, 1): 'Windows Server 2008 R2',
    (6, 2, 1): 'Windows 8',
    (6, 2): 'Windows Server 2012',
    (6, 3): 'Windows Server 2012 R2',
}
name_to_version = {value: key[:2] for key, value in version_to_name.items()}

class Windows(object): #pylint: disable-msg=R0902,R0904
    def __init__(self):
        from .interface import get_version_ex, get_system_info
        self.hyper_v = False
        self.architecture = ""
        self.service_pack = ""
        self.server_core = False
        self.edition = ""
        self.version = ""
        self._version_ex = get_version_ex()
        self._system_info = get_system_info()
        self._edition = ""
        self.analyze()

    def analyze(self):
        self.analyze_windows_version()
        self.analyze_windows_edition()
        self.analyze_windows_service_pack()
        self.analyze_windows_architecture()

    def analyze_windows_version(self):
        version = (self._version_ex.major_version, self._version_ex.minor_version, self._version_ex.product_type)
        self.version = version_to_name.get(version, version_to_name.get(version[:2], 'Unknown'))

    def analyze_windows_edition(self):
        self.server_core = False
        self.hyper_v = False
        self.edition = 'Standard'
        if self._version_ex.major_version == 5:
            self.analyze_windows5_edition()
        elif self._version_ex.major_version == 6:
            self._edition = self.get_windows6_edition()
            self.analyze_windows6_edition()
            if self._version_ex.minor_version >= 2:
                # http://msdn.microsoft.com/en-us/library/windows/desktop/ms724358(v=vs.85).aspx
                # PRODUCT_*_SERVER_CORE values are not returned in Windows Server 2012
                self.analyze_server_core_according_to_registry()
        else:
            self.edition = 'Unknown'

    def analyze_windows5_edition(self):
        from .constants import VER_SUITE_BACKOFFICE, VER_SUITE_COMPUTE_SERVER, VER_SUITE_DATACENTER
        from .constants import VER_SUITE_ENTERPRISE, VER_SUITE_STORAGE_SERVER, VER_SUITE_WH_SERVER
        (_, mask) = (self._version_ex.suite_mask, self._version_ex.product_type)
        if mask & VER_SUITE_BACKOFFICE:
            self.edition = 'BackOffice'
        elif mask & VER_SUITE_COMPUTE_SERVER:
            self.edition = 'Compute Cluster'
        elif mask & VER_SUITE_DATACENTER:
            self.edition = 'Datacenter'
        elif mask & VER_SUITE_ENTERPRISE:
            self.edition = 'Enterprise'
        elif mask & VER_SUITE_STORAGE_SERVER:
            self.edition = 'Storage'
        elif mask & VER_SUITE_WH_SERVER:
            self.edition = 'Home'
        else:
            self.edition = 'Standard'

    def get_windows6_edition(self):
        from .interface import get_product_info
        return get_product_info(self._version_ex.major_version,
                            self._version_ex.minor_version,
                            self._version_ex.service_pack_major,
                            self._version_ex.service_pack_minor)

    def analyze_server_core_according_to_dism(self):
        # http://stackoverflow.com/questions/13065479/how-to-detect-windows-2012-core-edition-c
        from os import environ, path
        from infi.execute import execute
        SYSTEMROOT = environ.get("SystemRoot", path.join("C:", "Windows"))
        dism = path.join(environ.get("SYSTEMROOT"), "System32", "dism.exe")
        pid = execute([dism, "/online", "/get-features", "/format:table"])
        self.server_core = any("ServerCore-FullServer" in line and "Disabled" in line for
                               line in pid.get_stdout().splitlines())

    def analyze_server_core_according_to_registry(self):
        # http://msdn.microsoft.com/en-us/library/windows/desktop/hh846315%28v=vs.85%29.aspx
        from infi.registry import LocalComputer
        from infi.registry.errors import AccessDeniedException
        KEY = r"SOFTWARE\Microsoft\Windows NT\CurrentVersion\Server\ServerLevels"
        FEATURES = ("ServerCore", "Server-Gui-Mgmt", "Server-Gui-Shell")
        try:
            store = LocalComputer().local_machine[KEY].values_store
        except AccessDeniedException:
            self.analyze_server_core_according_to_dism()
            return
        self.server_core = all(item in store for item in FEATURES)

    def analyze_windows6_edition(self):
        from .constants import PRODUCT_SUITE_CLUSTER, PRODUCT_SUITE_DATACENTER
        from .constants import PRODUCT_SUITE_ENTERPRISE, PRODUCT_SUITE_STORAGE, PRODUCT_SUITE_STANDARD
        from .constants import PRODUCT_SUITE_SMALL_BUSINESS
        from .constants import SERVER_CORE, HYPER_V
        if PRODUCT_SUITE_CLUSTER.__contains__(self._edition):
            self.edition = 'Compute Cluster'
        if PRODUCT_SUITE_DATACENTER.__contains__(self._edition):
            self.edition = 'Datacenter'
        elif PRODUCT_SUITE_ENTERPRISE.__contains__(self._edition):
            self.edition = 'Enterprise'
        elif PRODUCT_SUITE_STORAGE.__contains__(self._edition):
            self.edition = 'Storage'
        elif PRODUCT_SUITE_STANDARD.__contains__(self._edition):
            self.edition = 'Standard'
        elif PRODUCT_SUITE_SMALL_BUSINESS.__contains__(self._edition):
            self.edition = 'Small Business'
        else:
            self.edition = 'Client'
        if SERVER_CORE.__contains__(self._edition):
            self.server_core = True
        if HYPER_V.__contains__(self._edition):
            self.hyper_v = True

    def analyze_windows_service_pack(self):
        self.service_pack = self._version_ex.service_pack_major

    def analyze_windows_architecture(self):
        from .constants import PROCESSOR_ARCHITECTURE_AMD64, PROCESSOR_ARCHITECTURE_IA64, PROCESSOR_ARCHITECTURE_INTEL
        self.architecture = ''
        processor = self._system_info.processor_architecture
        if processor == PROCESSOR_ARCHITECTURE_AMD64:
            self.architecture = 'x64'
        elif processor == PROCESSOR_ARCHITECTURE_IA64:
            self.architecture = 'ia64'
        elif processor == PROCESSOR_ARCHITECTURE_INTEL:
            self.architecture = 'x86'
        else:
            self.architecture = 'unknown'

    def __str__(self):
        return "%s %s Service Pack %s" % (self.version, self.edition, self.service_pack)

    def is_windows_2000(self):
        return self.version == 'Windows 2000'

    def is_windows_xp(self):
        return self.version == 'Windows XP'

    def is_windows_2003(self):
        return self.version == 'Windows Server 2003'

    def is_windows_2008(self):
        return self.version == 'Windows Server 2008'

    def is_windows_2008_r2(self):
        return self.version == 'Windows Server 2008 R2'

    def is_windows_vista(self):
        return self.version == 'Windows Vista'

    def is_windows_7(self):
        return self.version == 'Windows 7'

    def is_windows_8(self):
        return self.version == 'Windows 8'

    def is_windows_2012(self):
        return self.version == 'Windows Server 2012'

    def is_windows_2012_r2(self):
        return self.version == 'Windows Server 2012 R2'

    def is_x86(self):
        return self.architecture == 'x86'

    def is_x64(self):
        return self.architecture == 'x64'

    def is_ia64(self):
        return self.architecture == 'ia64'

    def is_server_core(self):
        return self.server_core

    def is_hyper_v(self):
        return self.hyper_v

    def is_cluster_aware(self):
        if self.edition == 'Enterprise':
            return True
        if self.edition == 'Datacenter':
            return True
        return False

    def greater_than(self, os_name):
        if os_name not in name_to_version:
            raise Exception("Could not find OS name {}".format(os_name))
        actual_version = (self._version_ex.major_version, self._version_ex.minor_version)
        return actual_version > name_to_version[os_name]




#pylint: disable-msg=E1101
def print_records():
    from .interface import get_product_info, get_system_info, get_version_ex
    from .structures import OSVersionEx, SystemInfo

    version = get_version_ex()
    system_info = get_system_info()
    try:
        product_info = get_product_info(version.major_version, version.minor_version,
                                    version.service_pack_major, version.service_pack_minor)
    except: #pylint: disable-msg=W0702
        # product_info not available on Windows 2003, XP
        product_info = None
    print "OSVersionEx = %s" % repr(OSVersionEx.write_to_string(version))
    print "SystemInfo = %s" % repr(SystemInfo.write_to_string(system_info))
    print "ProductInfo = %s" % repr(product_info)
